from unittest.mock import MagicMock

import pytest
from jinja2 import Environment

from flask_restforms.main import FlaskRestForms, RestFormsJinja, RestFormsMiddleware


class AppMock(MagicMock):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.wsgi_app = MagicMock()
        self.jinja_options = {'extensions': []}
        self.jinja_env = MagicMock()


@pytest.mark.parametrize('request_method, field_value, expected_method', (
    ('POST', 'DELETE', 'DELETE'),
    ('POST', None, 'POST'),
    ('GET', '', 'GET'),
    ('POST', 'OTHER', 'OTHER'),
))
def test_middleware_updates_request_method(request_method, field_value, expected_method):
    method_field = 'method-field'
    app = AppMock()
    start_response = MagicMock()
    request = MagicMock()
    request.form = {method_field: field_value}
    environ = {
        'werkzeug.request': request,
        'REQUEST_METHOD': request_method,
    }
    middleware = RestFormsMiddleware(app.wsgi_app, method_field)

    response = middleware(environ, start_response)

    app.wsgi_app.assert_called_once_with(environ, start_response)
    assert response == app.wsgi_app.return_value
    assert environ['REQUEST_METHOD'] == expected_method


def test_flask_restforms_set_up():
    app = AppMock()

    FlaskRestForms(app)

    app.jinja_env.extend.assert_called_once_with(
        flask_restforms_field=app.config['REST_FORMS_METHOD_FIELD']
    )
    assert RestFormsJinja in app.jinja_options['extensions']
    assert isinstance(app.wsgi_app, RestFormsMiddleware)


def test_jinja_appends_hidden_input_to_forms():
    env = Environment()
    env.add_extension(RestFormsJinja)
    env.extend(flask_restforms_field='method_input')

    t = env.from_string('''
        <h1>Title</h1>
        <!-- <form>
            <button type="submit">Go</button>
        </form> -->
        <form action="/other/url/or/so">
            <button type="submit">Go</button>
        </form>
        <form method="GET">
            <button type="submit">Go</button>
        </form>
        <form method="DELETE">
            <button type="submit">Delete</button>
        </form>
        <form method="PATCH">
            <button type="submit">Edit</button>
        </form>
    ''')

    assert t.render() == '''
        <h1>Title</h1>
        <!-- <form>
            <button type="submit">Go</button>
        </form> -->
        <form action="/other/url/or/so">
            <button type="submit">Go</button>
        </form>
        <form method="GET">
            <button type="submit">Go</button>
        </form>
        <form method="POST"><input type="hidden" name="method_input" value="DELETE" />
            <button type="submit">Delete</button>
        </form>
        <form method="POST"><input type="hidden" name="method_input" value="PATCH" />
            <button type="submit">Edit</button>
        </form>
    '''



import re

from jinja2.ext import Extension
from jinja2.lexer import count_newlines
from jinja2.lexer import Token


class FlaskRestForms:
    DEFAULT_METHOD_FIELD = '__flask-restforms-method'

    def __init__(self, app):
        self.app = app
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        app.config.setdefault('REST_FORMS_METHOD_FIELD', self.DEFAULT_METHOD_FIELD)
        method_field = app.config['REST_FORMS_METHOD_FIELD']

        app.jinja_options['extensions'].append(RestFormsJinja)
        app.jinja_env.extend(flask_restforms_field=method_field)

        app.wsgi_app = RestFormsMiddleware(app.wsgi_app, method_field)


form_regex = re.compile(r'<form[^>]*>')
method_regex = re.compile(r'method=[\'\"]*(\w+)[\'\"]*')
class RestFormsJinja(Extension):
    IGNORED_METHODS = ['GET', 'HEAD']

    def filter_stream(self, stream):
        for token in stream:
            if token.type != 'data':
                yield token
                continue

            pos = 0
            lineno = token.lineno

            while 1:
                form_match = form_regex.search(token.value, pos)

                if form_match is None:
                    break

                new_pos = form_match.start()
                if new_pos > pos:
                    preval = token.value[pos:new_pos]
                    yield Token(lineno, 'data', preval)
                    lineno += count_newlines(preval)

                form_str = form_match.group()

                method_match = method_regex.search(form_str)
                method = method_match.group(1) if method_match else None

                if not method or method.upper() in self.IGNORED_METHODS:
                    val = token.value[new_pos:form_match.end()]
                    yield Token(lineno, 'data', val)
                    lineno += count_newlines(val)
                else:
                    method_str = method_match.group()
                    new_method = method_str.replace(method, 'POST')
                    new_form = form_str.replace(method_str, new_method)
                    yield Token(lineno, 'data', new_form)
                    lineno += count_newlines(new_form)

                    input_name = self.environment.flask_restforms_field
                    input_str = f'<input type="hidden" name="{input_name}" value="{method}" />'
                    yield Token(lineno, 'data', input_str)
                    lineno += count_newlines(input_str)

                pos = form_match.end()

            if pos < len(token.value):
                yield Token(lineno, 'data', token.value[pos:])


class RestFormsMiddleware:
    def __init__(self, app, method_field):
        self.app = app
        self.method_field = method_field

    def __call__(self, environ, start_response):
        request = self._get_request(environ)
        method = self._get_method_field_value(request)
        if method:
            environ['REQUEST_METHOD'] = method

        return self.app(environ, start_response)

    def _get_request(self, environ):
        return environ['werkzeug.request']

    def _get_method_field_value(self, request):
        return request.form.get(self.method_field)



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='Flask-RestForms',
    packages=['flask_restforms'],
    version='0.1.1',
    license='MIT',
    description='Add REST superpowers to your Jinja templates',
    author='Alan Munguia Cerda',
    url='https://www.github.com/alanmunguiacerda/flask-restforms/',
    keywords=['FLASK', 'JINJA', 'REST', 'MPA'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    python_requires=">= 3.6",
    install_requires=[
        'Flask',
        'Jinja2>=2.5'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Environment :: Web Environment',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ]
)